# Copyright (c) 2023 Animatea
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

__all__: typing.Sequence[str] = (
    "read",
    "getcwd",
    "join",
    "find_any",
)

import os
import typing


def read(filename: str) -> str:
    with open(filename) as file:
        return file.read()


def getcwd() -> str:
    return os.getcwd()


def join(*components: str) -> str:
    return os.path.join(*components)


def find_any(*filenames: str, base_dir: str = os.getcwd()) -> typing.List[str]:
    found_files: typing.List[str] = []

    for dirpath, dirnames, filenames_in_dir in os.walk(base_dir):
        for filename in filenames:
            if filename in filenames_in_dir:
                found_files.append(os.path.join(dirpath, filename))

    return found_files
